modeller
cccp  https://arteni.cs.dartmouth.edu/cccp/index.gen.php
IMP
ATSAS https://www.embl-hamburg.de/biosaxs/software.html
SAXSMoW https://saxs.ifsc.usp.br/
HHpred https://toolkit.tuebingen.mpg.de/hhpred
PSIPRED http://bioinf.cs.ucl.ac.uk/psipred/
DISOPRED http://bioinf.cs.ucl.ac.uk/psipred/?disopred=1
AllosMod https://salilab.org/allosmod
DomPred http://bioinf.cs.ucl.ac.uk/dompred
COILS/PCOILS https://toolkit.tuebingen.mpg.de/#/tools/pcoils
EMAN2 http://blake.bcm.edu/emanwiki/EMAN2
RELION https://www2.mrc-lmb.cam.ac.uk/relion/
SGD https://www.yeastgenome.org/
HeliQuest http://heliquest.ipmc.cnrs.fr/
Multicoil2 http://groups.csail.mit.edu/cb/multicoil2/cgi-bin/multicoil2.cgi
